@@ -71,12 +71,14 @@ extern "C" {
 #define SRL_RDR_BODY_POS_OFS_(buf) ((buf).pos - (buf).body_pos)
 #define srl_stack_push_and_set(_iter, _tag, _length, _stack_ptr) STMT_START {   \
     srl_stack_push_ptr((_iter)->pstack, (_stack_ptr));                          \
+    (_stack_ptr)->ridx = (_length);                                             \
+    (_stack_ptr)->length = (_length);                                           \
     (_stack_ptr)->offset = SRL_RDR_BODY_POS_OFS_((_iter)->buf);                 \
-    (_stack_ptr)->count = (_length);                                            \
-    (_stack_ptr)->idx = (_length);                                              \
+    (_stack_ptr)->prev_depth= 0;                                                \
     (_stack_ptr)->tag = (_tag);                                                 \
 } STMT_END
 
+#define SRL_ITER_STACK_PREALLOCATE (32)
 #define SRL_ITER_STACK_ROOT_TAG SRL_HDR_PACKET_START
 #define SRL_ITER_STACK_ON_ROOT(stack) ((stack)->ptr->tag == SRL_ITER_STACK_ROOT_TAG)
 
@@ -89,45 +91,70 @@ extern "C" {
 #define SRL_ITER_ERRORf3(fmt, var1, var2, var3) croak(SRL_ITER_BASE_ERROR_FORMAT fmt,  SRL_ITER_BASE_ERROR_ARGS, (var1), (var2), (var3))
 
 #ifdef TRACE_ITERATOR
-#   define SRL_ITER_TRACE(msg, args...)            SRL_RDR_TRACE("%s " msg, srl_debug_tabulator(iter), ## args)
+#   define SRL_ITER_TRACE(msg, args...)                                             \
+        SRL_RDR_TRACE("%s " msg, srl_debug_tabulator(iter), ## args)
+
+#   define SRL_ITER_TRACE_WITH_POSITION(msg, args...) STMT_START {                  \
+        SRL_RDR_TRACE(                                                              \
+            "%s " msg " (ofs %"UVuf" body_ofs %"UVuf")",                            \
+            srl_debug_tabulator(iter),                                              \
+            ## args,                                                                \
+            (UV) SRL_RDR_POS_OFS((iter)->pbuf),                                     \
+            (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf)                                 \
+        );                                                                          \
+    } STMT_END
+
 #   define SRL_ITER_REPORT_TAG(iter, tag) STMT_START {                              \
         SRL_RDR_TRACE(                                                              \
-            "%s tag SRL_HDR_%s (int: %d hex: 0x%x) at ofs %"UVuf" body_ofs %"UVuf,  \
+            "%s tag SRL_HDR_%s (hex: 0x%x) at ofs %"UVuf" body_ofs %"UVuf,          \
             srl_debug_tabulator((iter)),                                            \
             SRL_TAG_NAME((tag)),                                                    \
-            (tag), (tag),                                                           \
+            (tag),                                                                  \
             (UV) SRL_RDR_POS_OFS((iter)->pbuf),                                     \
             (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf)                                 \
         );                                                                          \
     } STMT_END
 #   define SRL_ITER_REPORT_STACK_STATE(iter) STMT_START {                           \
-        srl_iterator_stack_ptr stack_ptr = (iter)->pstack->ptr;                     \
-        SRL_RDR_TRACE(                                                              \
-            "%s stack state depth=%"IVdf" tag=SRL_HDR_%s (int: %d hex: 0x%x) "      \
-            "count=%u idx=%d offset=%"UVuf,                                         \
-            srl_debug_tabulator((iter)),                                            \
-            SRL_STACK_DEPTH((iter)->pstack),                                        \
-            SRL_TAG_NAME(stack_ptr->tag), stack_ptr->tag, stack_ptr->tag,           \
-            stack_ptr->count,                                                       \
-            stack_ptr->idx,                                                         \
-            stack_ptr->offset                                                       \
-        );                                                                          \
+        if (srl_stack_empty((iter)->pstack)) {                                      \
+            SRL_RDR_TRACE(                                                          \
+                "%s stack state depth=%"IVdf,                                       \
+                srl_debug_tabulator((iter)),                                        \
+                SRL_STACK_DEPTH((iter)->pstack)                                     \
+            );                                                                      \
+        } else {                                                                    \
+            srl_iterator_stack_ptr stack_ptr = (iter)->stack.ptr;                   \
+            SRL_RDR_TRACE(                                                          \
+                "%s stack state depth=%"IVdf" tag=SRL_HDR_%s (hex: 0x%x)"           \
+                " idx=%d (rdix=%d) offset=%"UVuf" length=%u",                       \
+                srl_debug_tabulator((iter)),                                        \
+                SRL_STACK_DEPTH((iter)->pstack),                                    \
+                SRL_TAG_NAME(stack_ptr->tag),                                       \
+                stack_ptr->tag,                                                     \
+                stack_ptr->length - stack_ptr->ridx,                                \
+                stack_ptr->ridx,                                                    \
+                stack_ptr->offset,                                                  \
+                stack_ptr->length                                                   \
+            );                                                                      \
+        }                                                                           \
     } STMT_END
 #else
 #   define SRL_ITER_TRACE(msg, args...)
+#   define SRL_ITER_TRACE_WITH_POSITION(msg, args...)
 #   define SRL_ITER_REPORT_TAG(iter, tag)
 #   define SRL_ITER_REPORT_STACK_STATE(iter)
 #endif
 
+// note that it's different from SRL_RDR_DONE()
+#define SRL_RDR_DONE_(buf) ((buf).pos > (buf).end)
 #define SRL_ITER_ASSERT_EOF(iter, msg) STMT_START {                                 \
-    if (expect_false(SRL_RDR_DONE((iter)->pbuf))) {                                 \
+    if (expect_false(SRL_RDR_DONE_((iter)->buf))) {                                 \
         SRL_RDR_ERROR_EOF((iter)->pbuf, (msg));                                     \
     }                                                                               \
 } STMT_END
 
 #define SRL_ITER_ASSERT_STACK(iter) STMT_START {                                    \
     assert(!srl_stack_empty((iter)->pstack));                                       \
-    if (expect_false(((int) (iter)->stack.ptr->idx)) < 0) {                         \
+    if (expect_false(((int) (iter)->stack.ptr->ridx)) < 0) {                        \
         SRL_ITER_ERROR("Stack is empty! Inconsistent state!");                      \
     }                                                                               \
 } STMT_END
@@ -157,7 +184,7 @@ extern "C" {
 } STMT_END
 
 /* function declaration */
-SRL_STATIC_INLINE void srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter);
+SRL_STATIC_INLINE void srl_iterator_rewind_stack_position(pTHX_ srl_iterator_t *iter);
 
 /* wrappers */
 UV srl_iterator_eof(pTHX_ srl_iterator_t *iter)     { return SRL_RDR_DONE(iter->pbuf) ? 1 : 0; }
@@ -188,7 +215,7 @@ srl_init_iterator(pTHX_ srl_iterator_t *iter, HV *opt)
 {
     assert(iter != NULL);
 
-    if (expect_false(srl_stack_init(aTHX_ &iter->stack, 32) != 0)) {
+    if (expect_false(srl_stack_init(aTHX_ &iter->stack, SRL_ITER_STACK_PREALLOCATE) != 0)) {
         Safefree(iter);
         croak("Out of memory");
     }
@@ -208,19 +235,20 @@ srl_init_iterator(pTHX_ srl_iterator_t *iter, HV *opt)
 }
 
 void
-srl_copy_iterator(pTHX_ srl_iterator_t *from, srl_iterator_t *to)
+srl_shallow_copy_iterator(pTHX_ srl_iterator_t *from, srl_iterator_t *to)
 {
+    srl_iterator_t *iter = from; // for SRL_ITER_TRACE
     assert(from != NULL);
     assert(to != NULL);
     SRL_ITER_TRACE("from=%p to=%p", from, to);
 
-    if (srl_stack_copy(aTHX_ &from->stack, &to->stack) != 0)
+    if (expect_false(srl_stack_copy(aTHX_ &from->stack, &to->stack)) != 0)
         croak("Out of memory");
 
     /* it's assumed that buf holds buffer owned by sv */
     to->document = from->document;
-    if (to->document) SvREFCNT_inc(to->document);
-    Copy(&from->buf, &to->buf, 1, srl_reader_buffer_t);
+    if (to->document) SvREFCNT_inc(to->document);       // shallow document copy
+    Copy(&from->buf, &to->buf, 1, srl_reader_buffer_t); // shallow buffer copy
 
     to->pstack = &to->stack;
     to->pbuf = &to->buf;
@@ -244,7 +272,7 @@ srl_deinit_iterator(pTHX_ srl_iterator_t *iter)
 void
 srl_destroy_iterator(pTHX_ srl_iterator_t *iter)
 {
-    srl_deinit_iterator(iter);
+    srl_deinit_iterator(aTHX_ iter);
     Safefree(iter);
 }
 
@@ -329,10 +357,10 @@ srl_iterator_reset(pTHX_ srl_iterator_t *iter)
         srl_stack_pop(stack); // does empty check internally
     }
 
-    srl_iterator_restore_stack_position(aTHX_ iter);
+    srl_iterator_rewind_stack_position(aTHX_ iter);
 }
 
-void
+IV
 srl_iterator_unite(pTHX_ srl_iterator_t *iter)
 {
     UV offset;
@@ -352,67 +380,79 @@ srl_iterator_unite(pTHX_ srl_iterator_t *iter)
     SRL_ITER_ASSERT_STACK(iter);
     iter->buf.pos = iter->buf.body_pos + offset;
 
+    SRL_ITER_TRACE_WITH_POSITION("after unite");
     SRL_ITER_REPORT_STACK_STATE(iter);
-    SRL_ITER_TRACE("ofs %"UVuf" body_ofs %"UVuf,
-                   (UV) SRL_RDR_POS_OFS((iter)->pbuf),
-                   (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf));
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
+    return iter->stack.depth;
 }
 
-void
+IV
 srl_iterator_disjoin(pTHX_ srl_iterator_t *iter)
 {
     srl_iterator_stack_ptr stack_ptr;
 
+    SRL_ITER_TRACE_WITH_POSITION("before disjoin");
     SRL_ITER_REPORT_STACK_STATE(iter);
-    SRL_ITER_TRACE("ofs %"UVuf" body_ofs %"UVuf,
-                   (UV) SRL_RDR_POS_OFS((iter)->pbuf),
-                   (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf));
 
-    // This record apart of being a boundary stores offset to idx's tag (i.e
+    // This record apart of being a boundary stores offset to ridx's tag (i.e
     // current tag). By default stack keeps offset to tag's starting point
     // (i.e. continer located in the buf).
 
     srl_stack_push_ptr(iter->pstack, stack_ptr);
     stack_ptr->offset = SRL_RDR_BODY_POS_OFS(iter->pbuf); // disjoint point
     stack_ptr->tag = SRL_ITER_STACK_ROOT_TAG;
-    stack_ptr->count = 1;
-    stack_ptr->idx = 1;
+    stack_ptr->length = 1;
+    stack_ptr->ridx = 1;
+
+    return iter->stack.depth;
 }
 
 
 SRL_STATIC_INLINE void
-srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter)
+srl_iterator_rewind_stack_position(pTHX_ srl_iterator_t *iter)
 {
     SRL_ITER_ASSERT_STACK(iter);
 
-    iter->stack.ptr->idx = iter->stack.ptr->count;
+    iter->stack.ptr->ridx = iter->stack.ptr->length;
     iter->buf.pos = iter->buf.body_pos + iter->stack.ptr->offset;
 
+    SRL_ITER_TRACE_WITH_POSITION("after rewind");
     SRL_ITER_REPORT_STACK_STATE(iter);
-    SRL_ITER_TRACE("ofs %"UVuf" body_ofs %"UVuf,
-                   (UV) SRL_RDR_POS_OFS((iter)->pbuf),
-                   (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf));
-
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 }
 
 #define srl_iterator_wrap_stack(iter, expected_depth, stack_ptr) STMT_START {                   \
-    SRL_ITER_TRACE("expected_depth=%"IVdf, (IV) (expected_depth));                              \
+    SRL_ITER_TRACE_WITH_POSITION("wrap_stack depth=%"IVdf, (IV) (expected_depth));              \
     SRL_ITER_REPORT_STACK_STATE(iter);                                                          \
                                                                                                 \
     assert((((IV) (expected_depth)) > iter->stack.depth) == 0);                                 \
                                                                                                 \
     while (    (iter)->stack.depth != (IV) (expected_depth)                                     \
-            && (stack_ptr)->idx == 0                                                            \
+            && (stack_ptr)->ridx == 0                                                           \
             && (stack_ptr)->tag != SRL_ITER_STACK_ROOT_TAG                                      \
           )                                                                                     \
     {                                                                                           \
         srl_stack_pop_nocheck((iter)->pstack);                                                  \
+        SRL_ITER_REPORT_STACK_STATE((iter));                                                    \
         (stack_ptr) = (iter)->stack.ptr;                                                        \
+                                                                                                \
+        if ((stack_ptr)->prev_depth) {                                                          \
+            iter->buf.pos = iter->buf.body_pos + (stack_ptr)->prev_depth;                       \
+            (stack_ptr)->prev_depth = 0;                                                        \
+                                                                                                \
+            /* Offset stored in prev_depth is offset of REFP tag. */                            \
+            /* At the same time, ridx was already decremented for REFP (see comments in */      \
+            /* srl_iterator_step_internal(). To compensate this, we increate ridx. */           \
+            (stack_ptr)->ridx++;                                                                \
+                                                                                                \
+            SRL_ITER_TRACE_WITH_POSITION("wrap_stack restore offset");                          \
+            DEBUG_ASSERT_RDR_SANE(iter->pbuf);                                                  \
+            SRL_ITER_REPORT_STACK_STATE(iter);                                                  \
+        }                                                                                       \
+                                                                                                \
         if (SRL_ITER_STACK_ON_ROOT(iter->pstack))                                               \
-            SRL_ITER_TRACE("root of stack reached");                                            \
+            SRL_ITER_TRACE_WITH_POSITION("root of stack reached");                              \
     }                                                                                           \
 } STMT_END
 
@@ -422,10 +462,14 @@ srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter)
     U8 tag;                                                                                     \
     UV length;                                                                                  \
                                                                                                 \
-    if (expect_false((stack_ptr)->idx == 0))                                                    \
+    if (expect_false((stack_ptr)->ridx == 0))                                                   \
         SRL_ITER_ERRORf1("Nothing to parse at depth=%"IVdf, SRL_STACK_DEPTH(iter->pstack));     \
                                                                                                 \
-    (stack_ptr)->idx--;                                                                         \
+    /* Iterator decrement ridx *before* parsing an element. This's done for simplicity. */      \
+    /* Also, it has some sense. For scalars there is no difference in when idx is decremented.*/\
+    /* For arrays/hashes/and objects think this way: element at current depth will be */        \
+    /* parsed as soon as iterator finished parsing element at depth+1. */                       \
+    (stack_ptr)->ridx--;                                                                        \
                                                                                                 \
     SRL_ITER_ASSERT_STACK(iter);                                                                \
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);                                                          \
@@ -440,7 +484,7 @@ read_again:
     iter->buf.pos++;                                                                            \
                                                                                                 \
     /* No code which decrease step, next or stack's counters should be added here.              \
-     * Otherwise the counters will be decreased twicer for tags like REFN, ALIAS, etc. */       \
+     * Otherwise the counters will be decreased twicer for tags like REFN, WEAKEN , etc. */     \
                                                                                                 \
     switch (tag & 0xE0) {                                                                       \
         case 0x0: /* POS_0 .. NEG_1 */                                                          \
@@ -484,7 +528,6 @@ read_again:
                     break;                                                                      \
                                                                                                 \
                 case SRL_HDR_REFN:                                                              \
-                case SRL_HDR_ALIAS:                                                             \
                 case SRL_HDR_WEAKEN:                                                            \
                     goto read_again;                                                            \
                                                                                                 \
@@ -501,6 +544,7 @@ read_again:
                                                                                                 \
                 case SRL_HDR_COPY:                                                              \
                 case SRL_HDR_REFP:                                                              \
+                case SRL_HDR_ALIAS:                                                             \
                     srl_skip_varint(aTHX_ iter->pbuf);                                          \
                     break;                                                                      \
                                                                                                 \
@@ -524,23 +568,104 @@ read_again:
 void
 srl_iterator_step_in(pTHX_ srl_iterator_t *iter, UV n)
 {
+    UV offset;
+    U8 tag, otag;
+    srl_reader_char_ptr orig_pos;
     srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_STACK(iter);
 
-    SRL_ITER_TRACE("n=%"UVuf, n);
+    SRL_ITER_TRACE_WITH_POSITION("n=%"UVuf, n);
     SRL_ITER_REPORT_STACK_STATE(iter);
 
     while (n--) {
         srl_iterator_wrap_stack(iter, -1, stack_ptr);
+
+        orig_pos = iter->buf.pos;
+        tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;
+        SRL_ITER_REPORT_TAG(iter, tag);
+        iter->buf.pos++;
+
+        if (tag == SRL_HDR_REFP || tag == SRL_HDR_ALIAS) {
+            /* store offset for REFP tag. Will be used in srl_iterator_wrap_stack() */
+            stack_ptr->prev_depth = orig_pos - iter->buf.body_pos; // almost same sa SRL_RDR_BODY_POS_OFS
+
+            offset = srl_read_varint_uv_offset(aTHX_ iter->pbuf, " while reading REFP tag");
+            iter->buf.pos = iter->buf.body_pos + offset;
+            DEBUG_ASSERT_RDR_SANE(iter->pbuf);
+
+            tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;
+            SRL_ITER_TRACE_WITH_POSITION("tag SRL_HDR_REFP points to tag SRL_HDR_%s",
+                                         SRL_TAG_NAME(tag));
+
+            switch (tag) {
+                case SRL_HDR_HASH:
+                case SRL_HDR_ARRAY:
+                    break;
+
+                case SRL_HDR_OBJECTV:
+                case SRL_HDR_OBJECTV_FREEZE:
+                case SRL_HDR_OBJECT:
+                case SRL_HDR_OBJECT_FREEZE:
+                    croak("not implemented OBJECT");
+
+                CASE_SRL_HDR_HASHREF:
+                CASE_SRL_HDR_ARRAYREF:
+                    croak("not implemente HASHREF ARRAYREF");
+
+                default:
+                    stack_ptr->prev_depth = 0;
+                    iter->buf.pos = orig_pos;
+                    break;
+            }
+        } else {
+            stack_ptr->prev_depth = 0;
+            iter->buf.pos = orig_pos;
+        }
+
         srl_iterator_step_internal(iter, stack_ptr);
     }
 
-    SRL_ITER_TRACE("Completed expected number of steps");
+    SRL_ITER_TRACE_WITH_POSITION("Completed expected number of steps");
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 }
 
+void
+srl_iterator_step_out(pTHX_ srl_iterator_t *iter, UV n)
+{
+    U32 idx;
+    IV depth = iter->stack.depth;
+    srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
+
+    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
+    SRL_ITER_ASSERT_STACK(iter);
+
+    SRL_ITER_TRACE_WITH_POSITION("n=%"UVuf, n);
+    SRL_ITER_REPORT_STACK_STATE(iter);
+
+    if (expect_false(n == 0)) return;
+
+    while (n--) {
+        if (expect_false(stack_ptr->tag == SRL_ITER_STACK_ROOT_TAG)) {
+            SRL_ITER_ERROR("Root of the stack is reached");
+        }
+
+        stack_ptr--;
+        depth--;
+    }
+
+    idx = stack_ptr->length - stack_ptr->ridx;
+    SRL_ITER_TRACE_WITH_POSITION("target depth: %"UVuf" idx=%u ridx=%d (length=%d at target depth)",
+                                 depth, idx, stack_ptr->ridx, stack_ptr->length);
+
+    assert(idx >= 0);
+    assert(idx <= stack_ptr->length);
+    assert(depth >= 0);
+
+    srl_iterator_until(aTHX_ iter, (UV) depth, idx);
+}
+
 /* srl_iterator_next() does N step on current stack.
  * It garantees that iterator remains on current stack level upon returing */
 
@@ -553,7 +678,7 @@ srl_iterator_next(pTHX_ srl_iterator_t *iter, UV n)
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_STACK(iter);
 
-    SRL_ITER_TRACE("n=%"UVuf, n);
+    SRL_ITER_TRACE_WITH_POSITION("n=%"UVuf, n);
     SRL_ITER_REPORT_STACK_STATE(iter);
 
     if (expect_false(n == 0)) return;
@@ -578,91 +703,89 @@ srl_iterator_next(pTHX_ srl_iterator_t *iter, UV n)
                           expected_depth, iter->stack.depth);
     }
 
-    SRL_ITER_TRACE("Did expected number of steps at depth %"IVdf, expected_depth);
+    SRL_ITER_TRACE_WITH_POSITION("Did expected number of steps at depth %"IVdf, expected_depth);
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 }
 
-/* srl_iterator_next_until_depth_and_idx() moves iterator forward until
- * expected stack level (depth) and index is reached. It can only go down the stack. */
+/* srl_iterator_until() moves iterator forward until expected stack level
+ * (depth) and index is reached. It can only go down the stack. */
 
 void
-srl_iterator_next_until_depth_and_idx(pTHX_ srl_iterator_t *iter, UV expected_depth, U32 expected_idx) {
+srl_iterator_until(pTHX_ srl_iterator_t *iter, UV depth, U32 idx) {
+    I32 ridx;
     IV current_depth = iter->stack.depth;
     srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_STACK(iter);
 
-    SRL_ITER_TRACE("expected_depth=%"UVuf" expected_idx=%u", expected_depth, expected_idx);
+    SRL_ITER_TRACE_WITH_POSITION("depth=%"UVuf" idx=%d", depth, idx);
     SRL_ITER_REPORT_STACK_STATE(iter);
 
-    if (expect_false((IV) expected_depth == current_depth && (IV) expected_idx == stack_ptr->idx))
+    if (expect_false((IV) depth > current_depth)) {
+        SRL_ITER_ERRORf2("srl_iterator_until() can only go forward, "
+                         "so depth=%"UVuf" should not be greater then current_depth=%"IVdf,
+                         depth, current_depth);
+    }
+
+    stack_ptr = iter->stack.begin + depth;
+    ridx = stack_ptr->length - idx;
+
+    if (expect_false((IV) ridx > stack_ptr->ridx)) {
+        SRL_ITER_ERRORf3("srl_iterator_until() can only go forward, "
+                         "so ridx=%u should not be greater then current "
+                         "index (%u) at depth=%"IVdf,
+                         ridx, stack_ptr->ridx, depth);
+    }
+
+    if (expect_false((IV) depth == current_depth && (IV) ridx == stack_ptr->ridx))
         return;
 
-    if (expect_false((IV) expected_depth > current_depth)) {
-        SRL_ITER_ERRORf2("srl_iterator_next_until_depth() can only go forward, "
-                         "so expected_depth=%"UVuf" should not be greater then current_depth=%"IVdf,
-                         expected_depth, current_depth);
-    }
-
-    stack_ptr = iter->stack.begin + expected_depth;
-    if (expect_false((IV) expected_idx > stack_ptr->idx)) {
-        SRL_ITER_ERRORf3("srl_iterator_next_until_depth() can only go forward, "
-                         "so expected_idx=%u should not be greater then current "
-                         "index (%u) at expected_depth=%"IVdf,
-                         expected_idx, stack_ptr->idx, expected_depth);
-    }
-
     stack_ptr = iter->stack.ptr;
 
     while (1) {
-        srl_iterator_wrap_stack(iter, expected_depth, stack_ptr);
+        srl_iterator_wrap_stack(iter, depth, stack_ptr);
 
-        if (iter->stack.depth == (IV) expected_depth) {
-            if (stack_ptr->idx == (IV) expected_idx)  break;
-            assert(((IV) expected_idx > stack_ptr->idx) == 0);
+        if (iter->stack.depth == (IV) depth) {
+            if (stack_ptr->ridx == (IV) ridx) break;
+            assert(((IV) ridx > stack_ptr->ridx) == 0);
         }
 
         srl_iterator_step_internal(iter, stack_ptr);
     }
 
-    assert(stack_ptr->idx == (IV) expected_idx);
-    assert(iter->stack.depth == (IV) expected_depth);
-    SRL_ITER_TRACE("Reached expected stack depth: %"UVuf " and idx: %u",
-                   expected_depth, expected_idx);
-
+    assert(stack_ptr->ridx == (IV) ridx);
+    assert(iter->stack.depth == (IV) depth);
+    SRL_ITER_TRACE_WITH_POSITION("Reached expected stack depth: %"UVuf " idx: %u ridx: %d", depth, idx, ridx);
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 }
 
 void
-srl_iterator_step_out(pTHX_ srl_iterator_t *iter, UV n)
+srl_iterator_rewind(pTHX_ srl_iterator_t *iter, UV n)
 {
     srl_stack_t *stack = iter->pstack;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_STACK(iter);
 
-    SRL_ITER_TRACE("n=%"UVuf, n);
+    SRL_ITER_TRACE_WITH_POSITION("n=%"UVuf, n);
     SRL_ITER_REPORT_STACK_STATE(iter);
 
-    // SRL_ITER_ASSERT_EOF(iter, "serialized object"); XXX need ability to go back on last element
-    // if (expect_false(n == 0)) return; XXX keep it as a feature?
-
     while (n--) {
         if (expect_false(SRL_ITER_STACK_ON_ROOT(stack))) {
-            SRL_ITER_ERROR("It was last object on stack, no more parents");
+            SRL_ITER_ERROR("Root of the stack is reached");
         }
 
         srl_stack_pop_nocheck(stack);
     }
 
-    srl_iterator_restore_stack_position(aTHX_ iter);
+    srl_iterator_rewind_stack_position(aTHX_ iter);
 }
 
 IV
 srl_iterator_array_goto(pTHX_ srl_iterator_t *iter, I32 idx)
 {
-    I32 s_idx;
+    I32 nidx, ridx;
     srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
@@ -670,40 +793,34 @@ srl_iterator_array_goto(pTHX_ srl_iterator_t *iter, I32 idx)
     SRL_ITER_ASSERT_STACK(iter);
     SRL_ITER_ASSERT_ARRAY_ON_STACK(iter);
 
-    SRL_ITER_TRACE("idx=%d", idx);
+    SRL_ITER_TRACE_WITH_POSITION("idx=%d", idx);
     SRL_ITER_REPORT_STACK_STATE(iter);
 
-    if (idx >= 0) {
-        s_idx = stack_ptr->count - idx;
-        if (idx >= (I32) stack_ptr->count) {
-            SRL_ITER_TRACE("Index is out of range, idx=%d count=%u", idx, stack_ptr->count);
-            return SRL_ITER_NOT_FOUND;
-        }
-    } else {
-        s_idx = -idx;
-        if (s_idx > (I32) stack_ptr->count) {
-            SRL_ITER_TRACE("Index is out of range, idx=%d count=%u", idx, stack_ptr->count);
-            return SRL_ITER_NOT_FOUND;
-        }
+    nidx = srl_iterator_normalize_idx(aTHX_ idx, stack_ptr->length);
+    if (nidx < 0 || nidx >= (I32) stack_ptr->length) {
+        SRL_ITER_TRACE("Index is out of range, idx=%d nidx=%d length=%u",
+                       idx, nidx, stack_ptr->length);
+        return SRL_ITER_NOT_FOUND;
     }
 
-    if (s_idx == stack_ptr->idx) {
+    ridx = stack_ptr->length - nidx;
+    if (ridx == stack_ptr->ridx) {
         return SRL_RDR_BODY_POS_OFS(iter->pbuf); // already at expected position
-    } else if (s_idx > stack_ptr->idx) {
-        SRL_ITER_ERRORf2("Can't go backwards, idx=%d, count=%u",
-                         idx, stack_ptr->count);
+    } else if (ridx > stack_ptr->ridx) {
+        SRL_ITER_ERRORf2("Can't go backwards, ridx=%d, length=%u",
+                         ridx, stack_ptr->length);
     }
 
     // srl_iterator_next garantee that we remans on current stack
-    srl_iterator_next(aTHX_ iter, stack_ptr->idx - s_idx);
-    assert(stack_ptr->idx == s_idx);
+    srl_iterator_next(aTHX_ iter, stack_ptr->ridx - ridx);
+    assert(stack_ptr->ridx == ridx);
     return SRL_RDR_BODY_POS_OFS(iter->pbuf);
 }
 
 IV
 srl_iterator_array_exists(pTHX_ srl_iterator_t *iter, I32 idx)
 {
-    I32 s_idx;
+    I32 nidx;
     srl_iterator_stack_ptr stack_ptr = iter->stack.ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
@@ -711,24 +828,17 @@ srl_iterator_array_exists(pTHX_ srl_iterator_t *iter, I32 idx)
     SRL_ITER_ASSERT_STACK(iter);
     SRL_ITER_ASSERT_ARRAY_ON_STACK(iter);
 
-    SRL_ITER_TRACE("idx=%d", idx);
+    SRL_ITER_TRACE_WITH_POSITION("idx=%d", idx);
     SRL_ITER_REPORT_STACK_STATE(iter);
 
-    if (idx >= 0) {
-        s_idx = stack_ptr->count - idx;
-        if (idx >= (I32) stack_ptr->count) {
-            SRL_ITER_TRACE("Index is out of range, idx=%d count=%u", idx, stack_ptr->count);
-            return SRL_ITER_NOT_FOUND;
-        }
-    } else {
-        s_idx = -idx;
-        if (s_idx > (I32) stack_ptr->count) {
-            SRL_ITER_TRACE("Index is out of range, idx=%d count=%u", idx, stack_ptr->count);
-            return SRL_ITER_NOT_FOUND;
-        }
+    nidx = srl_iterator_normalize_idx(aTHX_ idx, stack_ptr->length);
+    if (nidx < 0 || nidx >= (I32) stack_ptr->length) {
+        SRL_ITER_TRACE("Index is out of range, idx=%d nidx=%d length=%u",
+                       idx, nidx, stack_ptr->length);
+        return SRL_ITER_NOT_FOUND;
     }
 
-    return s_idx;
+    return nidx;
 }
 
 const char *
@@ -836,13 +946,13 @@ srl_iterator_hash_exists(pTHX_ srl_iterator_t *iter, const char *name, STRLEN na
     SRL_ITER_ASSERT_STACK(iter);
     SRL_ITER_ASSERT_HASH_ON_STACK(iter);
 
-    SRL_ITER_TRACE("name=%.*s", (int) name_len, name);
+    SRL_ITER_TRACE_WITH_POSITION("name=%.*s", (int) name_len, name);
     SRL_ITER_REPORT_STACK_STATE(iter);
 
-    while (stack_ptr->idx) {
-        stack_ptr->idx--; // do not make it be part of while clause
+    while (stack_ptr->ridx) {
+        stack_ptr->ridx--; // do not make it be part of while clause
         SRL_ITER_ASSERT_STACK(iter);
-        assert(stack_ptr->idx % 2 == 1);
+        assert(stack_ptr->ridx % 2 == 1);
         assert(iter->stack.depth == stack_depth);
         DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 
@@ -910,8 +1020,7 @@ srl_iterator_hash_exists(pTHX_ srl_iterator_t *iter, const char *name, STRLEN na
         if (   length == name_len
             && memcmp(name, key_ptr, name_len) == 0)
         {
-            SRL_ITER_TRACE("found key '%.*s' at offset %"UVuf,
-                         (int) name_len, name, SRL_RDR_BODY_POS_OFS(iter->pbuf));
+            SRL_ITER_TRACE_WITH_POSITION("found key '%.*s'", (int) name_len, name);
             return SRL_RDR_BODY_POS_OFS(iter->pbuf);
         }
 
@@ -1026,8 +1135,8 @@ UV
 srl_iterator_stack_index(pTHX_ srl_iterator_t *iter)
 {
     SRL_ITER_ASSERT_STACK(iter);
-    assert((I32) iter->stack.ptr->count >= iter->stack.ptr->idx); 
-    return (UV) (iter->stack.ptr->count - iter->stack.ptr->idx);
+    assert((I32) iter->stack.ptr->length >= iter->stack.ptr->ridx);
+    return (UV) (iter->stack.ptr->length - iter->stack.ptr->ridx);
 }
 
 UV
@@ -1038,7 +1147,7 @@ srl_iterator_stack_info(pTHX_ srl_iterator_t *iter, UV *length_ptr)
     srl_iterator_stack_ptr stack_ptr = stack->ptr;
 
     SRL_ITER_ASSERT_STACK(iter);
-    if (length_ptr) *length_ptr = stack_ptr->count;
+    if (length_ptr) *length_ptr = stack_ptr->length;
 
     switch (stack_ptr->tag) {
         case SRL_ITER_STACK_ROOT_TAG:
